"""A multithread-safe resource pool based on synchronized queue 
"""
# Copyright (c) 2002-2019 Aware Software, inc. All rights reserved.
# Copyright (c) 2005-2019 ikh software, inc. All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#
# queuepool/smtpcm.py - implements ConnectionManager for SMTP connections
#
import sys
import smtplib
import queuepool.pool as pool

def logg(*args, **kwargs):
   print(*args, file=sys.stderr, **kwargs)

class ConnectionManager(pool.ResourceManager):
   """ ConnectionManager for SMTP connections
   """
   def __init__(self, name, host='', port=0, user=None, password=None, *args, **kwargs):
      super().__init__(name)
      self.host = host
      self.port = port
      self.user = user
      self.password = password

   def __repr__(self):
      return str(dict(host=self.host, port=self.port, user=self.user, password='xxxxxxx', ResourceManager=super().__repr__()))

   def open(self):
      self.resource = smtplib.SMTP(self.host, self.port)
      self.resource.starttls()
      self.resource.login(self.user, self.password)
      super().open()

   def close(self):
      self.resource.quit()
      self.resource = None
      super().close()

   def takeRepair(self):
      super().takeRepair()

   def putRepair(self):
      super().putRepair()

   def sendmail(self, *args, **kwargs):
      self.resource.sendmail(*args, **kwargs)






"""A multithread-safe resource pool based on synchronized queue 
"""
# Copyright (c) 2002-2019 Aware Software, inc. All rights reserved.
# Copyright (c) 2005-2019 ikh software, inc. All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#
# queuepool/psycopg2cm.py - implements ConnectionManager for psycopg2 connections
#
import sys
import json
import psycopg2 as pg
from psycopg2 import sql as pgsql
from psycopg2 import extensions as _ext
import queuepool.pool as pool

def logg(*args, **kwargs):
   print(*args, file=sys.stderr, **kwargs)

class ConnectionManager(pool.ResourceManager):
   """ ConnectionManager for psycopg2 connections
   """
   def __init__(self, name, isolation_level=None, readonly=None, deferrable=None, autocommit=None, *args, **kwargs):
      super().__init__(name)
      self.autocommit = autocommit
      self.isolation_level = isolation_level
      self.readonly = readonly
      self.deferrable = deferrable
      self._args = args
      self._kwargs = kwargs

   def __repr__(self):
      return str(dict(autocommit=self.autocommit, isolation_level=self.isolation_level, readonly=self.readonly, deferrable=self.deferrable, ResourceManager=super().__repr__()))

   def open(self):
      self.resource = pg.connect(*self._args, **self._kwargs)
      self.resource.set_session(isolation_level=self.isolation_level, readonly=self.readonly, deferrable=self.deferrable, autocommit=self.autocommit)
      super().open()

   def close(self):
      if not self.resource.closed:
         self.resource.close()
      self.resource = None
      super().close()

   def repair(self):
      if self._isOpen:
         if not self.resource.closed:
            s = self.resource.info.transaction_status
            if s == _ext.TRANSACTION_STATUS_UNKNOWN:
               # server connection lost
               self.close()
               #logg(f"ConnectionManager: repair: server connection lost: {self}")
            elif s != _ext.TRANSACTION_STATUS_IDLE:
               # connection in error or in transaction (ACTIVE, INTRANS, INERROR)
               self.resource.rollback()
               #logg(f"ConnectionManager: repair: still in transaction: {self}")
            else:
               # regular idle connection
               pass
         else:
            #logg(f"ConnectionManager: repair: connection was closed outside of resource manager: {self}")
            self.close()

   def takeRepair(self):
      super().takeRepair()

   def putRepair(self):
      self.repair()
      super().putRepair()


class ConnectionManagerExtended(ConnectionManager):
   def __init__(self, *args, **kwargs):
      super().__init__(*args, **kwargs)

   def now_expr(self, t=0): # t - seconds
      return f"((extract(epoch from now())+{t})*1000)::bigint"

   def query(self, query, args=None):
      with self.resource.cursor() as c:
         c.execute(query, args)
         ds = c.description
         xs = c.fetchall()
      rs = list(map( lambda z: dict( map( lambda x,y: (x[0],y), ds, z)), xs ))
      return rs

   def querys(self, query, args=None):
      with self.resource.cursor() as c:
         c.execute(query, args)
         ds = c.description
         xs = c.fetchall()
      return xs,ds

   def query1(self, query, args=None):
      with self.resource.cursor() as c:
         c.execute(query, args)
         ds = c.description
         xs = c.fetchall()
      rs = list(map( lambda z: dict( map( lambda x,y: (x[0],y), ds, z)), xs ))
      return rs[0] if len(rs)==1 else None

   def qexec(self, query, args=None):
      with self.resource.cursor() as c:
         c.execute(query, args)

   def geojson(self, geomColumn, attrColumns, selectQuery, args=None, prec=5):
      q = pgsql.SQL("with x as ("+selectQuery+") select ST_AsGeoJSON({},"+str(int(prec))+") as geometry,{} from x").format(pgsql.Identifier(geomColumn),pgsql.SQL(', ').join(map(pgsql.Identifier, attrColumns)))
      rs = self.query(q, args)
      fs = list(map(lambda r: dict(type="Feature", geometry=json.loads(r['geometry']), properties={k: r[k] for k in filter(lambda k: k!='geometry', r)}), rs))
      res = dict(type="FeatureCollection", features=fs)
      return res





"""A multithread-safe resource pool based on synchronized queue 
"""
# Copyright (c) 2002-2019 Aware Software, inc. All rights reserved.
# Copyright (c) 2005-2019 ikh software, inc. All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#
# queuepool/pool.py - implements Pool and ResourceManager
#
import sys
import queue
from datetime import datetime
import threading
import time

def logg(*args, **kwargs):
   print(*args, file=sys.stderr, **kwargs)

class ResourceManager:
   """Abstract ResourceManager
   """
   def __init__(self, name):
      self.name = name
      self.resource = None
      self._pool = None
      self._isOpen = False
      self._lastOpened = None
      self._lastUsed = None
      self._usageCount = None
      #logg(f"ResourceManager: initialized resource {self!r}")

   def __str__(self):
      return self.name

   def __repr__(self):
      return str(dict(name=self.name, _isOpen=self._isOpen, _lastOpened=self._lastOpened, _lastUsed=self._lastUsed, _usageCount=self._usageCount))

   def open(self):
      t = datetime.now()
      self._isOpen = True
      self._lastOpened = t
      self._lastUsed = t
      self._usageCount = 0
      #logg(f"ResourceManager: opened resource {self!r}")

   def close(self):
      self._isOpen = False
      #logg(f"ResourceManager: closed resource {self!r}")

   def takeRepair(self):
      pass

   def putRepair(self):
      pass

   def __enter__(self):
      return self

   def __exit__(self, exc_type, exc_value, traceback):
      if self._pool.closeOnException and exc_type is not None:
         self.close()
      self._pool.put(self)
      return False

class Pool:
   """Multithread-safe resource pool based on synchronized queue
   """
   def __init__(self, name, capacity, maxIdleTime=300.0, maxOpenTime=300.0, maxUsageCount=1000, closeOnException=True):
      self.name = name
      self.capacity = capacity
      self._maxIdleTime = maxIdleTime
      self._maxOpenTime = maxOpenTime
      self._maxUsageCount = maxUsageCount
      self.closeOnException = closeOnException
      self._queue = queue.LifoQueue(self.capacity)
      self._recyclerThread = None
      #logg(f"Pool: initialized pool {self!r}")

   def __str__(self):
      return self.name

   def __repr__(self):
      return str(dict(name=self.name, capacity=self.capacity, _maxIdleTime=self._maxIdleTime, _maxOpenTime=self._maxOpenTime, _maxUsageCount=self._maxUsageCount, closeOnException=self.closeOnException))

   def take(self):
      t = datetime.now()
      r = self._queue.get()
      r._pool = self
      self._recycle(r)
      r.takeRepair()
      if not r._isOpen:
         r.open()
      r._usageCount += 1
      r._lastUsed = t
      #logg(f"Pool '{self}': took resource {r!r}")
      return r

   def put(self,r):
      t = datetime.now()
      r._pool = self
      r._lastUsed = t
      self._recycle(r)
      r.putRepair()
      self._queue.put(r)
      #logg(f"Pool '{self}': put resource {r!r}")

   def _recycle(self,r):
      t = datetime.now()
      if r._isOpen:
         isMaxIdle = self._maxIdleTime is not None and (t - r._lastUsed).total_seconds() > self._maxIdleTime
         isMaxOpen = self._maxOpenTime is not None and (t - r._lastOpened).total_seconds() > self._maxOpenTime
         isMaxUsage = self._maxUsageCount is not None and r._usageCount >= self._maxUsageCount
         if isMaxIdle or isMaxOpen or isMaxUsage:
               r.close()
               #logg(f"Pool '{self}': recycled resource {r!r}, {dict(isMaxIdle=isMaxIdle, isMaxOpen=isMaxOpen, isMaxUsage=isMaxUsage)}")

   def recycle(self):
      done = False
      rs = []
      opened0 = 0
      opened1 = 0
      while not done and len(rs) != self.capacity:
         try:
            r = self._queue.get(block=False)
            opened0 += 1 if r._isOpen else 0
            self._recycle(r)
            opened1 += 1 if r._isOpen else 0
            rs.append(r)
         except queue.Empty as e:
            done = True
      for r in reversed(rs):
         self._queue.put(r)
      #logg(f"Pool '{self}': recycled {len(rs)} resources: opened: {opened0} -> {opened1}")
      
   def startRecycler(self, interval=60):
      if self._recyclerThread is None:
         self._recyclerInterval = interval
         self._recyclerThread = threading.Thread(target=self._recyclerLoop, name='queuepool recycler', daemon=True)
         self._recyclerThread.start()
      
   def _recyclerLoop(self):
      while True:
         time.sleep(self._recyclerInterval)
         self.recycle()
   



# Copyright (c) 2002-2019 Aware Software, inc. All rights reserved.
# Copyright (c) 2005-2019 ikh software, inc. All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#
# queuepool/psycopg2cm_tests.py - tests for psycopg2 connections
#

import sys
from psycopg2 import extensions as _ext
from psycopg2 import errors as _errors
import time
import multiprocessing.pool as mpp
import random
import queuepool.pool as qp
import queuepool.psycopg2cm as pgcm

def logg(*args, **kwargs):
   print(*args, file=sys.stderr, **kwargs)

def test1(host, dbname,user, password):
   pool = qp.Pool(name=dbname, capacity=20, maxIdleTime=60, maxOpenTime=600, maxUsageCount=1000, closeOnException=True)
   for i in range(pool.capacity):
      pool.put(pgcm.ConnectionManager(name=dbname+'-'+str(i), autocommit=False, isolation_level=_ext.ISOLATION_LEVEL_SERIALIZABLE, host=host, dbname=dbname, user=user, password=password))
   pool.startRecycler(5)
   
   # with context manager
   with pool.take() as cm:
      conn = cm.resource
      with conn.cursor() as c:
         c.execute("select count(*) from accounts")
         rs = c.fetchall() + [conn.get_backend_pid()]
         logg('Result 1: ', rs)
         #raise Exception('bum!')

   # without context manager
   connmgr = pool.take()
   try:
      conn = connmgr.resource
      with conn.cursor() as c:
         c.execute("select count(*) from accounts")
         rs = c.fetchall() + [conn.get_backend_pid()]
         logg('Result 2: ', rs)
         #raise Exception('bam!')
   except Exception as e:
      if pool.closeOnException:
         connmgr.close()
      raise
   finally:
      pool.put(connmgr)

   # example with thread pool
   def fetch(dbpool, name):
      res = []
      for i in range(10):
         with dbpool.take() as cm:
            conn = cm.resource
            with conn: # transaction
               with conn.cursor() as c:
                  done = False
                  while not done:
                     try:
                        c.execute("select x from x")
                        #time.sleep(0.1)
                        (x,), = c.fetchall()
                        x = x + 1
                        c.execute("update x set x=%s", (x,))
                        res.append(x)
                        done = True
                     except _errors.SerializationFailure as e:
                        conn.rollback()
                        logg('SerializationFailure: ', name, i, x)
                        pass
                     if not done:
                        time.sleep(random.random() * 0.01) 
                  #logg(name, i, x)
            with conn.cursor() as c:
               c.execute("select * from x")
               # simulate still in transaction
      # simulate conn closed outside of conn manager
      connm = dbpool.take()
      connm.resource.close()
      dbpool.put(connm)
      return res

   tpool = mpp.ThreadPool(2)
   rs = tpool.starmap(fetch, [(pool, 'task-'+str(x)) for x in range(100)])
   logg(sorted(sum(rs,[])))
   #time.sleep(10)




import setuptools

with open("README.rst", "r") as f:
    long_description = f.read()

setuptools.setup(
   name="queuepool",
   version="1.3.2",
   author="ikh software, inc.",
   author_email="ikh@ikhsoftware.com",
   description="A multithread-safe resource pool based on synchronized queue",
   long_description=long_description,
   long_description_content_type="text/x-rst",
   url="https://github.com/ikhomyakov/queuepool",
   packages=setuptools.find_packages(),
   classifiers=[
      'Programming Language :: Python :: 3',
      'License :: OSI Approved :: BSD License',
      'Operating System :: OS Independent',
   ],
   python_requires = '>= 3.7',
   install_requires=[
      'psycopg2 >= 2.8.2',
   ],
   project_urls={
        'Bug Reports': 'https://github.com/ikhomyakov/queuepool/issues',
   },
)

